

class UnsupportedFileFormatError(Exception):
    """
    This exception is intended to communicate that the file extension is not one of
    the supported file types and cannot be parsed with AICSImage.
    """

    def __init__(self, type_, **kwargs):
        super().__init__(**kwargs)
        self.type_ = type_

    def __str__(self):
        return f"AICSImage module does not support this image file type: '{self.type_}'"


class InvalidDimensionOrderingError(Exception):
    """
    A general exception that can be thrown when handling dimension ordering or validation. Should be provided a message
    for the user to be given more context.
    """

    def __init__(self, message: str, **kwargs):
        super().__init__(**kwargs)
        self.message = message

    def __str__(self):
        return self.message


class MultiSceneCziException(Exception):
    """
    This exception is intended to be thrown when a CZI file has multiple scenes. This is only to
    be thrown if the Reader is given a multi-scene CZI files and the backend library isn't able
    to read multi-scene CZI.
    """
    pass


class ConflictingArgumentsError(Exception):
    """
    This exception is returned when 2 arguments to the same function are in conflict.
    """
    pass



import logging
import typing
from typing import Optional, Type

import numpy as np

from . import constants, transforms, types
from .exceptions import InvalidDimensionOrderingError, UnsupportedFileFormatError
from .readers import CziReader, DefaultReader, NdArrayReader, OmeTiffReader, TiffReader
from .readers.reader import Reader

log = logging.getLogger(__name__)


class AICSImage:
    """
    AICSImage takes microscopy image data types (files / bytestreams) of varying dimensions ("ZYX", "TCZYX", "CYX") and
    puts them into a consistent 6D "STCZYX" ordered numpy.ndarray. The data, metadata are lazy loaded and can be
    accessed as needed. Note the dims are assumed to match "STCZYX" from right to left meaning if 4 dimensional
    data is provided then the dimensions are assigned to be "CZYX", 2 dimensional would be "YX". This guessed assignment
    is only for file types without dimension metadata (ie not .ome.tiff or .czi).

    Note: if you absolutely know the dims and they are not as guessed then you can override them by immediately setting
    them after initialization see example below.

    Simple Example
    --------------
    with open("filename.czi", 'rb') as fp:
        img = AICSImage(fp)
        data = img.data  # data is a 6D "STCZYX" object
        metadata = img.metadata  # metadata from the file, an xml.etree
        zstack_t8 = img.get_image_data("ZYX", S=0, T=8, C=0)  # returns a 3D "ZYX" numpy.ndarray

    zstack_t10 = data[0, 10, 0, :, :, :]  # access the S=0, T=10, C=0 "ZYX" cube

    File Examples
    -------------

    with AICSImage("filename.ome.tif") as img:
        # do work with img
        data = img.data
        metadata = img.metadata
        # img will be closed automatically at end of scope

    OmeTif
        img = AICSImage("filename.ome.tif")
        img.close()
    CZI (Zeiss)
        img = AICSImage("filename.czi") or AICSImage("filename.czi", max_workers=8)
        img.close()
    Tiff
        img = AICSImage("filename.tif")
        img.close()
    Png/Gif/...
        img = AICSImage("filename.png")
        img.close()
        img = AICSImage("filename.gif")
        img.close()

    Bytestream Examples
    -------------------
    OmeTif
        with open("filename.ome.tif", 'rb') as fp:
            img = AICSImage(fp)
    CZI
        with open("filename.czi", 'rb') as fp:
            img = AICSImage(fp, max_workers=7)
    Tiff/Png/Gif
        with open("filename.png", 'rb') as fp:
            img = AICSImage(fp)

    Numpy.ndarray Example
    ---------------------
    blank = numpy.zeros((2, 600, 900))
    img = AICSImage(blank)

    Example with known dimensions different than guessed
    ----------------------------------------------------
    img = AICSImage('TCXimage.gif')
    img.reader.dims = 'TCX'
    data = img.data  # get a 6D ndarray back in "STCZYX" order
    """

    SUPPORTED_READERS = [CziReader, OmeTiffReader, TiffReader, DefaultReader]

    def __init__(
        self,
        data: typing.Union[types.FileLike, np.ndarray],
        known_dims: Optional[str] = None,
        **kwargs,
    ):
        """
        Constructor for AICSImage class intended for providing a unified interface for dealing with
        microscopy images. To extend support to a new reader simply add a new reader child class of
        Reader ([readers/reader.py]) and add the class to SUPPORTED_READERS in AICSImage.

        Parameters
        ----------
        data: String with path to ometif/czi/tif/png/gif file, or ndarray with up to 6 dimensions
        known_dims: Optional string with the known dimension order. If None, the reader will attempt to parse dim order.
        kwargs: Parameters to be passed through to the reader class
                       max_workers (optional Czi) specifies the number of worker threads for the backend library
        """

        # Hold onto known dims until data is requested
        self._known_dims = known_dims

        # Dims should nearly always be default dim order unless explictly overridden
        self.dims = constants.DEFAULT_DIMENSION_ORDER

        # Lazy load later
        self._data = None
        self._metadata = None

        # Determine reader class and load data
        reader_class = self.determine_reader(data)
        self._reader = reader_class(data, **kwargs)

    @staticmethod
    def determine_reader(data: types.ImageLike) -> Type[Reader]:
        """Cheaply check to see if a given file is a recognized type.
        Currently recognized types are TIFF, OME TIFF, and CZI.
        If the file is a TIFF, then the description (OME XML if it is OME TIFF) can be retrieved via read_description.
        Similarly, if the file is a CZI, then the metadata XML can be retrieved via read_description.
        """
        # The order of the readers in this list is important.
        # Example: if TiffReader was placed before OmeTiffReader, we would never use the OmeTiffReader.
        for reader_class in [
            NdArrayReader,
            CziReader,
            OmeTiffReader,
            TiffReader,
            DefaultReader,
        ]:
            if reader_class.is_this_type(data):
                return reader_class

        raise UnsupportedFileFormatError(type(data))

    @property
    def data(self):
        """
        Returns
        -------
        Returns a numpy.ndarray with dimension ordering "STCZYX"
        """
        if self._data is None:
            reader_data = self._reader.data

            # Read and reshape and handle delayed known dims reshape
            self._data = transforms.reshape_data(
                data=reader_data,
                given_dims=self._known_dims or self.reader.dims,
                return_dims=self.dims,
            )
        return self._data

    def size(self, dims: str = "STCZYX"):
        """
        Parameters
        ----------
        dims: A string containing a list of dimensions being requested.  The default is to return the 6 standard dims

        Returns
        -------
        Returns a tuple with the requested dimensions filled in
        """
        dims = dims.upper()
        if not (all(d in "STCZYX" for d in dims)):
            raise InvalidDimensionOrderingError(f"Invalid dimensions requested: {dims}")
        if not (all(d in self.dims for d in dims)):
            raise InvalidDimensionOrderingError(f"Invalid dimensions requested: {dims}")
        return tuple([self.data.shape[self.dims.index(dim)] for dim in dims])

    @property
    def size_x(self):
        """
        Returns
        -------
        Returns the x size
        """
        return self.size("X")[0]

    @property
    def size_y(self):
        """
        Returns
        -------
        Returns the y size
        """
        return self.size("Y")[0]

    @property
    def size_z(self):
        """
        Returns
        -------
        Returns the z size
        """
        return self.size("Z")[0]

    @property
    def size_c(self):
        """
        Returns
        -------
        Returns the c size
        """
        return self.size("C")[0]

    @property
    def size_t(self):
        """
        Returns
        -------
        Returns the t size
        """
        return self.size("T")[0]

    @property
    def size_s(self):
        """
        Returns
        -------
        Returns the s size
        """
        return self.size("S")[0]

    @property
    def metadata(self):
        """
        Returns
        -------
        The Metadata from the Czi, or Ome.Tiff file, or other base class type with metadata.
        For pure image files and empty string or None is returned.

        """
        if self._metadata is None:
            self._metadata = self._reader.metadata
        return self._metadata

    @property
    def reader(self) -> Reader:
        """
        This property returns the class created to read the image file type.
        The intent is that if the AICSImage class doesn't provide a raw enough
        interface then the base class can be used directly.

        Returns
        -------
        A child of Reader, CziReader OmeTiffReader, TiffReader, DefaultReader, etc.

        """
        return self._reader

    def get_image_data(
        self, out_orientation: str = None, copy: bool = False, **kwargs
    ) -> np.ndarray:
        """

        Parameters
        ----------
        out_orientation: A string containing the dimension ordering desired for the returned ndarray
        copy: boolean value to get image data by reference or by value [True, False]

        kwargs:
            C=1: specifies Channel 1
            T=3: specifies the fourth index in T
            D=n: D is Dimension letter and n is the index desired D should not be present in the out_orientation

        Returns
        -------
        ndarray with dimension ordering that was specified with out_orientation
        Note: if a requested dimension is not present in the data the dimension is added with
        a depth of 1. The default return dimensions are "STCZYX".
        """
        out_orientation = self.dims if out_orientation is None else out_orientation
        if out_orientation == self.dims:
            return self.data
        return transforms.reshape_data(
            data=self.data,
            given_dims=self.dims,
            return_dims=out_orientation,
            copy=copy,
            **kwargs,
        )

    def get_channel_names(self, scene: int = 0):
        """

        Parameters
        ----------
        scene: the index of the scene for which to return channel names

        Returns
        -------
        list of strings representing the channel names
        """
        try:
            names = self._reader.get_channel_names(scene)
        except AttributeError:
            names = [str(i) for i in range(self.size_c)]
        return names

    def close(self):
        self.reader.close()

    def __repr__(self) -> str:
        return f"<AICSImage [{type(self.reader).__name__}]>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def imread(data: types.ImageLike, **kwargs) -> np.ndarray:
    with AICSImage(data, **kwargs) as image:
        imagedata = image.data
    return imagedata



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io

import imageio
import numpy as np

from .. import exceptions
from .reader import Reader


class DefaultReader(Reader):
    """
    DefaultReader('file.jpg')

    A catch all for image file reading that uses imageio as its back end.

    Parameters
    ----------
    file: types.FileLike
        A path or bytes object to read from.

    Notes
    -----
    Because this is simply a wrapper around imageio, no metadata is returned. However, dimension order is returned with
    dimensions assumed in order but with extra dimensions removed depending on image shape.

    This reader always returns the first YX plane of an image. If you need to read GIFs for example, please use
    `imageio.mimread` to get a list of YXC planes for the frames of the GIF.

    For more details about how `imageio.imread`, differs from `imageio.volread` and `imageio.mimread`, please refer to
    their documentation.
    """

    @property
    def data(self) -> np.ndarray:
        if self._data is None:
            self._data = imageio.imread(self._bytes)

        return self._data

    @property
    def dims(self) -> str:
        """
        `imageio.imread` returns the first YX plane of an image, except in the case where the image is RGB / RGBA, in
        which case it returns the first YXC plane of an image.

        This dims property handles those cases by making assumptions about the dimension order based off the shape of
        the data stored in the reader. In the case where more data than an YXC plane is returned, it uses the
        `guess_dim_order` function. However, I can't tell when that should ever get run and is really just used as a
        safety catch all.
        """
        if self._dims is None:
            if len(self.data.shape) == 2:
                self._dims = "YX"
            elif len(self.data.shape) == 3:
                self._dims = "YXC"
            else:
                self._dims = self.guess_dim_order(self.data.shape)

        return self._dims

    @dims.setter
    def dims(self, dims: str):
        # Check amount of provided dims against data shape
        if len(dims) != len(self.data.shape):
            raise exceptions.InvalidDimensionOrderingError(
                f"Provided too many dimensions for the associated file. "
                f"Received {len(dims)} dimensions [dims: {dims}] "
                f"for image with {len(self.data.shape)} dimensions [shape: {self.data.shape}]."
            )

        # Set the dims
        self._dims = dims

    @property
    def metadata(self) -> None:
        return None

    @staticmethod
    def _is_this_type(buffer: io.BufferedIOBase) -> bool:
        try:
            imageio.get_reader(buffer)
            return True
        except ValueError:
            return False



from .czi_reader import CziReader  # noqa: F401
from .default_reader import DefaultReader  # noqa: F401
from .ndarray_reader import NdArrayReader  # noqa: F401
from .ome_tiff_reader import OmeTiffReader  # noqa: F401
from .tiff_reader import TiffReader  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from .. import exceptions
from .reader import Reader


class NdArrayReader(Reader):
    """
    NdArrayReader(np.ones((1,2,3)))

    A catch all for numpy ndarray reading.

    Parameters
    ----------
    arr: numpy.ndarray
        An in memory numpy ndarray.

    Notes
    -----
    Because this is simply a wrapper around numpy ndarray, no metadata is returned. However, dimension order is
    returned with dimensions assumed in order but with extra dimensions removed depending on image shape.
    """

    def __init__(self, arr: np.ndarray):
        self._data = arr
        self._dims = self.guess_dim_order(self.data.shape)

    @property
    def data(self) -> np.ndarray:
        return self._data

    @property
    def dims(self) -> str:
        return self._dims

    @dims.setter
    def dims(self, dims: str):
        # Check amount of provided dims against data shape
        if len(dims) != len(self.data.shape):
            raise exceptions.InvalidDimensionOrderingError(
                f"Provided too many dimensions for the associated array. "
                f"Received {len(dims)} dimensions [dims: {dims}] "
                f"for image with {len(self.data.shape)} dimensions [shape: {self.data.shape}]."
            )

        # Set the dims
        self._dims = dims

    @property
    def metadata(self) -> None:
        return None

    @staticmethod
    def _is_this_type(arr: np.ndarray) -> bool:
        return isinstance(arr, np.ndarray)

    def close(self) -> None:
        pass



import os

import imageio
import numpy as np


class PngWriter:
    """This class can take 3D arrays of CYX pixel values and writes them to a png

    Example:
        image = numpy.ndarray([3, 1024, 2048])
        # There needs to be some sort of data inside the image array
        writer = png_writer.PngWriter("file.png")
        writer.save(image)

        image2 = numpy.ndarray([3, 1024, 2048])
        # There needs to be some sort of data inside the image2 array
        with png_writer.PngWriter("file2.png") as writer2:
            writer2.save(image2)
    """

    def __init__(self, file_path, overwrite_file=None):
        """
        Class initializer
        :param file_path: path to image output location
        :param overwrite_file: flag to overwrite image or pass over image if it already exists
            None : (default) throw IOError if file exists
            True : overwrite existing file if file exists
            False: silently perform no write actions if file exists
        """
        self.file_path = file_path
        self.silent_pass = False
        if os.path.isfile(self.file_path):
            if overwrite_file:
                os.remove(self.file_path)
            elif overwrite_file is None:
                raise IOError("File {} exists but user has chosen not to overwrite it".format(self.file_path))
            elif overwrite_file is False:
                self.silent_pass = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        pass

    def save(self, data):
        """Takes in an array of CYX pixel values and writes them to a png

        :param data: a CYX or YX array with C being the rgb channels for each pixel value
        """
        if self.silent_pass:
            return

        # check for rgb, rgba, or r
        if len(data.shape) == 3:
            assert data.shape[0] in [4, 3, 2, 1]
            # if three dimensions, transpose to YXC (imsave() needs it in these axes)
            data = np.transpose(data, (2, 1, 0))
            # if there's only one channel, repeat across the next two channels
            if data.shape[2] == 1:
                data = np.repeat(data, repeats=3, axis=2)
            elif data.shape[2] == 2:
                data = np.pad(data, ((0, 0), (0, 0), (0, 1)), 'constant')
        elif len(data.shape) != 2:
            raise ValueError("Data was not of dimensions CYX or YX")

        imageio.imwrite(self.file_path, data, format="png")

    def save_slice(self, data, z=0, c=0, t=0):
        """Exactly the same functionality as save() but allows the interface to be the same as OmeTiffWriter

        :param data: a CYX or YX array with C being the rgb channels for each pixel value
        :param z: an arbitrary z index that does nothing
        :param c: an arbitrary c index that does nothing
        :param t: an arbitrary t index that does nothing
        """
        if self.silent_pass:
            return

        self.save(data)



import io
from typing import Optional

import numpy as np
import tifffile

from .. import types
from ..buffer_reader import BufferReader
from . import reader


class TiffReader(reader.Reader):
    """This class is used to open and process the contents of a generic tiff file.

    The load function will get a 3D ZYX array from a tiff file.
    """

    @staticmethod
    def _is_this_type(buffer: io.BufferedIOBase) -> bool:
        with BufferReader(buffer) as buffer_reader:
            # Per the TIFF-6 spec (https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf),
            # 'II' is little-endian (Intel format) and 'MM' is big-endian (Motorola format)
            if buffer_reader.endianness not in [buffer_reader.INTEL_ENDIAN, buffer_reader.MOTOROLA_ENDIAN]:
                return False
            magic = buffer_reader.read_uint16()

            # Per TIFF-6, magic is 42.
            if magic == 42:
                ifd_offset = buffer_reader.read_uint32()
                if ifd_offset == 0:
                    return False

            # Per BigTIFF (https://www.awaresystems.be/imaging/tiff/bigtiff.html), magic is 43.
            if magic == 43:
                # Alex magic here...
                if buffer_reader.read_uint16() != 8:
                    return False
                if buffer_reader.read_uint16() != 0:
                    return False
                ifd_offset = buffer_reader.read_uint64()
                if ifd_offset == 0:
                    return False
            return True

    @staticmethod
    def get_image_description(buffer: io.BufferedIOBase) -> Optional[bytearray]:
        """Retrieve the image description as one large string."""
        description_length = 0
        description_offset = 0

        with BufferReader(buffer) as buffer_reader:
            # Per the TIFF-6 spec (https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf),
            # 'II' is little-endian (Intel format) and 'MM' is big-endian (Motorola format)
            if buffer_reader.endianness not in [buffer_reader.INTEL_ENDIAN, buffer_reader.MOTOROLA_ENDIAN]:
                return None
            magic = buffer_reader.read_uint16()

            # Per TIFF-6, magic is 42.
            if magic == 42:
                found = False
                while not found:
                    ifd_offset = buffer_reader.read_uint32()
                    if ifd_offset == 0:
                        return None
                    buffer_reader.buffer.seek(ifd_offset, 0)
                    entries = buffer_reader.read_uint16()
                    for n in range(0, entries):
                        tag = buffer_reader.read_uint16()
                        type = buffer_reader.read_uint16()
                        count = buffer_reader.read_uint32()
                        offset = buffer_reader.read_uint32()
                        if tag == 270:
                            description_length = count - 1  # drop the NUL from the end
                            description_offset = offset
                            found = True
                            break

            # Per BigTIFF (https://www.awaresystems.be/imaging/tiff/bigtiff.html), magic is 43.
            if magic == 43:
                # Alex magic here...
                if buffer_reader.read_uint16() != 8:
                    return None
                if buffer_reader.read_uint16() != 0:
                    return None
                found = False
                while not found:
                    ifd_offset = buffer_reader.read_uint64()
                    if ifd_offset == 0:
                        return None
                    buffer_reader.buffer.seek(ifd_offset, 0)
                    entries = buffer_reader.read_uint64()
                    for n in range(0, entries):
                        tag = buffer_reader.read_uint16()
                        type = buffer_reader.read_uint16()  # noqa: F841
                        count = buffer_reader.read_uint64()
                        offset = buffer_reader.read_uint64()
                        if tag == 270:
                            description_length = count - 1  # drop the NUL from the end
                            description_offset = offset
                            found = True
                            break

            if description_offset == 0:
                # Nothing was found
                return bytearray("")
            else:
                buffer_reader.buffer.seek(description_offset, 0)
                return bytearray(buffer_reader.buffer.read(description_length))

    def __init__(self, file: types.FileLike, **kwargs):
        super().__init__(file, **kwargs)
        self.tiff = tifffile.TiffFile(self._bytes)

    def close(self):
        self.tiff.close()
        super().close()

    def dtype(self):
        return self.tiff.pages[0].dtype

    @property
    def data(self) -> np.ndarray:
        if self._data is None:
            self._data = self.tiff.asarray()
        return self._data

    @property
    def dims(self) -> str:
        if self._dims is None:
            self._dims = self.guess_dim_order(self.data.shape)

        return self._dims

    @dims.setter
    def dims(self, value: str) -> None:
        self._dims = value

    @property
    def metadata(self) -> str:
        if self._metadata is None:
            description = self.get_image_description(self._bytes)
            if description is None:
                self._metadata = ''
            else:
                self._metadata = description.decode()
        return self._metadata



from .aics_image import AICSImage  # noqa: F401
from .aics_image import imread  # noqa: F401

# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "3.0.7"



from abc import ABC, abstractmethod

from .. import constants, types


class Writer(ABC):
    """Write STCZYX data arrays to file, with accompanying metadata
    Will overwrite existing files of same name.

    Example:
        1) construct and use as object
        image = numpy.ndarray([1, 10, 3, 1024, 2048])
        writer = DerivedWriter("file.ome.tif")
        writer.save(image)
        writer.close()

        2) use the "with" pattern
        image2 = numpy.ndarray([5, 486, 210])
        # There needs to be some sort of data inside the image2 array
        with DerivedWriter("file2.ome.tif") as writer2:
            write2.set_metadata(myMetaData)
            writer2.save(image2)
    """

    def __init__(self, file_path: types.PathLike):
        """
        Class initializer
        :param file_path: path to image output location
        """
        self.file_path = file_path

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @abstractmethod
    def close(self) -> None:
        """Close file objects"""
        pass

    @abstractmethod
    def save(self, data, dims=constants.DEFAULT_DIMENSION_ORDER, **kwargs) -> None:
        """Write to open file"""
        pass



import io
import logging
import warnings
from typing import Optional, Tuple
from xml.etree import ElementTree

import numpy as np

from aicsimageio import types

from ..buffer_reader import BufferReader
from ..exceptions import MultiSceneCziException, UnsupportedFileFormatError
from .reader import Reader

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from ..vendor import czifile

log = logging.getLogger(__name__)


class CziReader(Reader):
    """
    CziReader is intended for reading single scene Czi files. It is meant to handle the specifics of using the backend
    library to create a unified interface. This enables higher level functions to duck type the File Readers.
    """
    ZEISS_2BYTE = b'ZI'             # First two characters of a czi file according to Zeiss docs
    ZEISS_10BYTE = b'ZISRAWFILE'    # First 10 characters of a well formatted czi file.

    def __init__(self, file: types.FileLike, max_workers: Optional[int] = None, **kwargs):
        """

        Parameters
        ----------
        file : a file like object ("Filename.czi", Path("/path/Filename.czi") or an open stream to the data
        max_workers : (Optional) the number of cores the backend library is allowed to use to load the data in the file.
        """
        super().__init__(file, **kwargs)
        try:
            self.czi = czifile.CziFile(self._bytes)
        except Exception:
            log.error("czifile could not parse this input")
            raise UnsupportedFileFormatError("exception from with CziFile backend library.")

        if self._is_multiscene():
            raise MultiSceneCziException(
                "File is Multiscene. The backend library CziFile can only read single scene images."
            )

        self._max_workers = max_workers

    @staticmethod
    def _is_this_type(buffer: io.BufferedIOBase) -> bool:
        with BufferReader(buffer) as buffer_reader:
            if buffer_reader.endianness != CziReader.ZEISS_2BYTE:
                return False
            header = buffer_reader.endianness + buffer_reader.read_bytes(8)
            return header == CziReader.ZEISS_10BYTE

    @property
    def data(self) -> np.ndarray:
        """
        Returns
        -------
        the data from the czi file with the native order (i.e. "TZCYX")
        """
        if self._data is None:
            # load the data
            self._data = self.czi.asarray(max_workers=self._max_workers)
        return self._data

    @property
    def dims(self) -> str:
        """
        Returns
        -------
        The native shape of the image.
        """
        return self.czi.axes

    @property
    def metadata(self) -> ElementTree:
        """
        Lazy load the metadata from the CZI file
        Returns
        -------
        The xml Element Tree of the metadata
        """
        if self._metadata is None:
            # load the metadata
            self._metadata = self.czi.metadata
        return self._metadata

    def close(self):
        """
        Close the czi file handle and perform any upstream cleanup
        Returns
        -------
        None
        """
        self.czi.close()
        super().close()

    def dtype(self) -> np.dtype:
        """
        Returns
        -------
        the data type of the ndarray being returned (uint16, uint8, etc)
        """
        return self.czi.dtype

    def size_s(self):
        """
        Returns
        -------
        The number of scenes in the data
        """
        return self._size_of_dimension('S')

    def size_z(self):
        """
        Returns
        -------
        The number of Z slices in the stack
        """
        return self._size_of_dimension('Z')

    def size_c(self):
        """
        Returns
        -------
        The number of Channels present in the data
        """
        return self._size_of_dimension('C')

    def size_t(self):
        """
        Returns
        -------
        The number of time steps in the data
        """
        return self._size_of_dimension('T')

    def size_x(self):
        """
        Returns
        -------
        The number of pixels in the images X axis
        """
        return self._size_of_dimension('X')

    def size_y(self):
        """
        Returns
        -------
        The number of pixels in the images Y axis
        """
        return self._size_of_dimension('Y')

    def _size_of_dimension(self, dimension: str) -> int:
        """
        Parameters
        ----------
        dimension : str (a single character)

        Raises
        ------
        If a string of length greater or smaller than 1 is passed in raise a TypeError

        Returns
        -------
        The size of the dimension in the data, if the dimension is not found in the "BTCZYX" type string
        then the default dimension size of 1 is returned.
        """

        index = self._lookup_dimension_index(dimension)
        if index == -1:
            return 1
        return self.czi.shape[index]

    def _lookup_dimension_index(self, dimension: str) -> int:
        """
        Use the axes metadata in the czi file to find the dimension index, additionally this
        function should be used for determining if a Dimension is present in the native data.
        Parameters
        ----------
        dimension : str (a single character)
            sensible values are any one of ('V', 'H', 'M', 'B', 'I', 'S', 'T', 'R', 'Z', 'C', 'Y', 'X')
            most likely values are one of ('S', 'T', 'C', 'Z', 'Y', 'X')

        Raises
        ------
        If a string of length greater or smaller than 1 is passed in raise a TypeError

        Returns
        -------
        the integer position of the channel or -1 if the character is not present in the file description
        """
        if len(dimension) != 1:
            raise TypeError(f"channel lookup requested with channel {dimension}")
        return self.czi.axes.find(dimension)

    def _is_multiscene(self):
        """
        Check if the metadata the czi is multiscene

        Returns
        -------
        True if multi-scene, False if single-scene
        """
        index = self._lookup_dimension_index('S')
        if index < 0:
            return False
        img_shape = self.czi.filtered_subblock_directory[0].shape
        return img_shape[index] != 1

    def get_channel_names(self, scene: int = 0):
        chelem = self.metadata.findall("./Metadata/Information/Image/Dimensions/Channels/Channel")
        return [ch.get("Name") for ch in chelem]

    # TODO refactor this utility function into a metadata wrapper class
    def _getmetadataxmltext(self, findpath, default=None):
        ref = self.metadata.find(findpath)
        if ref is None:
            return default
        return ref.text

    def get_physical_pixel_size(self, scene: int = 0) -> Tuple[float]:
        px = float(self._getmetadataxmltext("./Metadata/Scaling/Items/Distance[@Id='X']/Value", "1.0"))
        py = float(self._getmetadataxmltext("./Metadata/Scaling/Items/Distance[@Id='Y']/Value", "1.0"))
        pz = float(self._getmetadataxmltext("./Metadata/Scaling/Items/Distance[@Id='Z']/Value", "1.0"))
        return (px, py, pz)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Collection of types used across multiple objects and functions.
"""

from io import BufferedIOBase
from pathlib import Path
from typing import Union, NamedTuple, Any

import numpy as np

# Imaging Data Types
SixDArray = np.ndarray  # In order STCZYX

# IO Types
PathLike = Union[str, Path]
BufferLike = Union[bytes, BufferedIOBase]
FileLike = Union[PathLike, BufferLike]
ImageLike = Union[FileLike, np.ndarray]


class LoadResults(NamedTuple):
    data: SixDArray
    dims: str
    metadata: Any



import logging
import warnings
from collections import Counter

import numpy as np

from .exceptions import ConflictingArgumentsError

log = logging.getLogger(__name__)


def reshape_data(data: np.ndarray, given_dims: str, return_dims: str, copy: bool = False, **kwargs) -> np.ndarray:
    """
    Reshape the data into return_dims, pad missing dimensions, and prune extra dimensions.
    Warns the user to use the base reader if the depth of the Dimension being removed is not 1.

    Parameters
    ----------
    data: a numpy.ndarray of arbitrary shape but with the dimensions specified in given_dims
    given_dims: the dimension ordering of data, "CZYX", "VBTCXZY" etc
    return_dims: the dimension ordering of the return data
    copy: True or False, if true copy the data object if false modify the object/view into the object
    kwargs:
        C=1 => desired specific channel, if C in the input data has depth 3 then C=1 returns the 2nd slice (0 indexed)
        Z=10 => desired specific channel, if Z in the input data has depth 20 then Z=10 returns the 11th slice
    Returns
    -------
    a numpy.ndarray in return_dims order, if return_dims=DEFAULT_DIMS then the return would have order "STCZYX"

    """
    # copy the data object if copy=True is in kwargs
    data = data.copy() if copy else data
    # check for conflicts like return_dims='TCZYX' and fixed channels 'C=1'
    for dim in return_dims:
        if kwargs.get(dim) is not None:
            msg = f"Argument return_dims={return_dims} and argument {dim}={kwargs.get(dim)} conflict. Check usage."
            raise ConflictingArgumentsError(msg)

    # add each dimension not included in original data
    new_dims = given_dims
    excluded_dims = "".join(set(return_dims) - set(given_dims))
    for dim in excluded_dims:
        data = np.expand_dims(data, axis=0)
        new_dims = dim + new_dims  # add the missing Dimension to the front

    # if given dims contains a Dimension not in DEFAULT_DIMS and its depth is 1 remove it
    # if it's larger than 1 give a warning and suggest interfacing with the Reader object
    extra_dims = "".join(set(given_dims) - set(return_dims))
    for dim in extra_dims:
        index = new_dims.find(dim)
        if data.shape[index] > 1:
            index_depth = kwargs.get(dim)
            if index_depth is None:
                msg = (f'Data has dimension {dim} with depth {data.shape[index]}, assuming {dim}=0 is  '
                       f'the desired value, if not the case specify {dim}=x where '
                       f'x is an integer in [0, {data.shape[index]}).')
                warnings.warn(msg=msg, category=UserWarning, stacklevel=2)
                index_depth = 0
            if index_depth >= data.shape[index]:
                raise IndexError(f'Dimension specified with {dim}={index_depth} '
                                 f'but Dimension shape is {data.shape[index]}.')
            planes = np.split(data, data.shape[index], axis=index)  # split dim into list of arrays
            data = planes[index_depth]  # take the specified value of the dim
        data = np.squeeze(data, axis=index)  # remove the dim from ndarray
        new_dims = new_dims[0:index] + new_dims[index + 1:]  # clip out the Dimension from new_dims
    # any extra dimensions have been removed, only a problem if the depth is > 1
    return transpose_to_dims(data, given_dims=new_dims, return_dims=return_dims)  # don't pass kwargs or 2 copies


def transpose_to_dims(data: np.ndarray, given_dims: str, return_dims: str, copy: bool = False) -> np.ndarray:
    """
    This shuffles the data dimensions from given_dims to return_dims. Each dimension must be present in
    given_dims must be used in return_dims

    Parameters
    ----------
    data: the input data with dimensions given_dims
    given_dims: the dimensions of data
    return_dims: the reordered set of given_dims to return
    copy: if True copy the data object passed in if false modify the numpy.ndarray

    Returns
    -------
    a numpy.ndarray with known_dims

    """
    # copy the data object if copy=True is in kwargs
    data = data.copy() if copy else data

    # Use a counter to track that the contents are composed of the same letters and that no letter is repeated
    if Counter(given_dims) != Counter(return_dims) or max(Counter(given_dims).values()) > 1:
        raise ConflictingArgumentsError(f"given_dims={given_dims} and return_dims={return_dims} are incompatible.")
    # resort the data into return_dims order
    match_map = {dim: given_dims.find(dim) for dim in given_dims}
    transposer = []
    for dim in return_dims:
        if match_map[dim] == -1:
            raise ConflictingArgumentsError(f'Dimension {dim} requested but not present in given_dims={given_dims}.')
        transposer.append(match_map[dim])
    data = data.transpose(transposer)
    return data



from . import types


class BufferReader:

    INTEL_ENDIAN = b'II'
    MOTOROLA_ENDIAN = b'MM'

    def __init__(self, buffer: types.FileLike):
        self.buffer = buffer
        self.previous_position = self.buffer.tell()
        self.description_length = 0
        self.description_offset = 0
        self.endianness = None

    def __enter__(self):
        self.buffer.seek(0)
        self.endianness = bytearray(self.buffer.read(2))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    def reset(self):
        self.buffer.seek(self.previous_position)

    # All of these read_uint* routines obey the endianness, with 'II' being little-endian
    # and 'MM' being big-endian (per TIFF-6)
    def read_uint16(self):
        value = bytearray(self.buffer.read(2))
        return (value[0] + (value[1] << 8)) if self.endianness == self.INTEL_ENDIAN else (value[1] + (value[0] << 8))

    def read_uint32(self):
        value = bytearray(self.buffer.read(4))
        if self.endianness == self.INTEL_ENDIAN:
            return value[0] + (value[1] << 8) + (value[2] << 16) + (value[3] << 24)
        return value[3] + (value[2] << 8) + (value[1] << 16) + (value[0] << 24)

    def read_uint64(self):
        if self.endianness == self.INTEL_ENDIAN:
            return self.read_uint32() + (self.read_uint32() << 32)
        return (self.read_uint32() << 32) + self.read_uint32()

    def read_bytes(self, n_bytes: int):
        return bytearray(self.buffer.read(n_bytes))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import numpy as np

from .. import constants, types


class Reader(ABC):

    _bytes = None

    _data = None
    _dims = None
    _metadata = None

    def __init__(self, file: types.FileLike, **kwargs):
        # Convert to BytesIO
        self._bytes = self.convert_to_buffer(file)

    @staticmethod
    def guess_dim_order(shape: tuple) -> str:
        return constants.DEFAULT_DIMENSION_ORDER[len(constants.DEFAULT_DIMENSION_ORDER) - len(shape):]

    @staticmethod
    def convert_to_buffer(file: types.FileLike) -> io.BufferedIOBase:
        # Check path
        if isinstance(file, (str, Path)):
            # This will both fully expand and enforce that the filepath exists
            f = Path(file).expanduser().resolve(strict=True)

            # This will check if the above enforced filepath is a directory
            if f.is_dir():
                raise IsADirectoryError(f)

            return open(f, "rb")

        # Convert bytes
        elif isinstance(file, bytes):
            return io.BytesIO(file)

        # Set bytes
        elif isinstance(file, io.BytesIO):
            return file

        # Special case for ndarray because already in memory
        elif isinstance(file, np.ndarray):
            return file

        # Raise
        else:
            raise TypeError(
                f"Reader only accepts types: [str, pathlib.Path, bytes, io.BytesIO], received: {type(file)}"
            )

    @classmethod
    def is_this_type(cls, file: types.FileLike) -> bool:
        buffer = cls.convert_to_buffer(file)
        return cls._is_this_type(buffer)

    @staticmethod
    @abstractmethod
    def _is_this_type(buffer: io.BufferedIOBase) -> bool:
        pass

    @property
    @abstractmethod
    def data(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def dims(self) -> str:
        pass

    @property
    @abstractmethod
    def metadata(self) -> Any:
        pass

    def load(self) -> types.LoadResults:
        return types.LoadResults(self.data, self.dims, self.metadata)

    def close(self) -> None:
        self._bytes.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from __future__ import print_function

import os

import numpy as np
import tifffile

from aicsimageio.vendor import omexml
from aicsimageio.exceptions import InvalidDimensionOrderingError

BYTE_BOUNDARY = 2 ** 21

# This is the threshold to use BigTiff, if it's the 4GB boundary it should be 2**22 but
# the libtiff writer was unable to handle a 2GB numpy array. It would be great if we better
# understood exactly what this threshold is and how to calculate it but for now this is a
# stopgap working value


class OmeTiffWriter:
    """This class can take arrays of pixel values and do the necessary metadata creation to write them
    properly in OME xml format.

    Example:
        image = numpy.ndarray([1, 10, 3, 1024, 2048])
        # There needs to be some sort of data inside the image array
        writer = ome_tiff_writer.OmeTiffWriter("file.ome.tif")
        writer.save(image)

        image2 = numpy.ndarray([5, 486, 210])
        # There needs to be some sort of data inside the image2 array
        with ome_tiff_writer.OmeTiffWriter("file2.ome.tif") as writer2:
            writer2.save(image2)

        # Convert a CZI file into OME Tif.
        reader = cziReader.CziReader("file3.czi")
        writer = ome_tiff_writer.OmeTiffWriter("file3.ome.tif")
        writer.save(reader.load())

    """

    def __init__(self, file_path, overwrite_file=None):
        """
        Class initializer
        :param file_path: path to image output location
        :param overwrite_file: flag to overwrite image or pass over image if it already exists
            None : (default) throw IOError if file exists
            True : overwrite existing file if file exists
            False: silently perform no write actions if file exists
        """
        self.file_path = file_path
        self.omeMetadata = omexml.OMEXML()
        self.silent_pass = False
        if os.path.isfile(self.file_path):
            if overwrite_file:
                os.remove(self.file_path)
            elif overwrite_file is None:
                raise IOError("File {} exists but user has chosen not to overwrite it".format(self.file_path))
            elif overwrite_file is False:
                self.silent_pass = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        pass

    def save(self, data, ome_xml=None, channel_names=None, image_name="IMAGE0", pixels_physical_size=None,
             channel_colors=None, dimension_order="STZCYX"):
        """
        Save an image with the proper OME xml metadata.

        Parameters
        ----------
        data: An array of 3, 4, or 5 dimensions to be written out to a file.
        ome_xml:
        channel_names: The names for each channel to be put into the OME metadata
        image_name: The name of the image to be put into the OME metadata
        pixels_physical_size: The physical size of each pixel in the image
        channel_colors: The channel colors to be put into the OME metadata
        dimension_order: The dimension ordering in the data array.  Will be assumed STZCYX if not specified

        Returns
        -------

        """
        if self.silent_pass:
            return

        shape = data.shape
        ndims = len(shape)

        assert (ndims == 5 or ndims == 4 or ndims == 3), "Expected 3, 4, or 5 dimensions in data array"

        # assert valid characters in dimension_order
        if not (all(d in "STCZYX" for d in dimension_order)):
            raise InvalidDimensionOrderingError(f"Invalid dimension_order {dimension_order}")
        if (dimension_order[-2:] != "YX"):
            raise InvalidDimensionOrderingError(f"Last two characters of dimension_order {dimension_order} expected to \
                be YX.  Please transpose your data.")
        if (len(dimension_order) < ndims):
            raise InvalidDimensionOrderingError(f"dimension_order {dimension_order} must have at least as many \
                dimensions as data shape {shape}")
        if (dimension_order.find("S") > 0):
            raise InvalidDimensionOrderingError(f"S must be the leading dim in dimension_order {dimension_order}")
        # todo ensure no letter appears more than once?

        # ensure dimension_order is same len as shape
        if len(dimension_order) > ndims:
            dimension_order = dimension_order[-ndims:]

        # if this is 3d data, then expand to 5D and add appropriate dimensions
        if ndims == 3:
            data = np.expand_dims(data, axis=0)
            data = np.expand_dims(data, axis=0)
            # prepend either TC, TZ or CZ
            if dimension_order[0] == 'T':
                dimension_order = "CZ" + dimension_order
            elif dimension_order[0] == 'C':
                dimension_order = "TZ" + dimension_order
            elif dimension_order[0] == 'Z':
                dimension_order = "TC" + dimension_order

        # if this is 4d data, then expand to 5D and add appropriate dimensions
        elif ndims == 4:
            data = np.expand_dims(data, axis=0)
            # prepend either T, C, or Z
            first2 = dimension_order[:2]
            if first2 == "TC" or first2 == "CT":
                dimension_order = "Z" + dimension_order
            elif first2 == "TZ" or first2 == "ZT":
                dimension_order = "C" + dimension_order
            elif first2 == "CZ" or first2 == "ZC":
                dimension_order = "T" + dimension_order

        if ome_xml is None:
            self._make_meta(data, channel_names=channel_names, image_name=image_name,
                            pixels_physical_size=pixels_physical_size, channel_colors=channel_colors,
                            dimension_order=dimension_order)
        else:
            pixels = ome_xml.image().Pixels
            pixels.populate_TiffData()
            self.omeMetadata = ome_xml
        xml = self.omeMetadata.to_xml().encode()

        tif = tifffile.TiffWriter(self.file_path, bigtiff=self._size_of_ndarray(data=data) > BYTE_BOUNDARY)

        # check data shape for TZCYX or ZCYX or ZYX
        if ndims == 5 or ndims == 4 or ndims == 3:
            # minisblack instructs TiffWriter to not try to infer rgb color within the data array
            # metadata param fixes the double image description bug
            tif.save(data, compress=9, description=xml, photometric='minisblack', metadata=None)

        tif.close()

    def save_slice(self, data, z=0, c=0, t=0):
        """ this doesn't do the necessary functionality at this point

        TODO:
            * make this insert a YX slice in between two other slices inside a full image
            * data should be a 5 dim array

        :param data:
        :param z:
        :param c:
        :param t:
        :return:
        """
        if self.silent_pass:
            return

        assert len(data.shape) == 2
        assert data.shape[0] == self.size_y()
        assert data.shape[1] == self.size_x()
        tif = tifffile.TiffWriter(self.file_path)
        tif.save(data, compress=9)
        tif.close()

    def set_metadata(self, ome_metadata):
        self.omeMetadata = ome_metadata

    def size_z(self):
        return self.omeMetadata.image().Pixels.SizeZ

    def size_c(self):
        return self.omeMetadata.image().Pixels.SizeC

    def size_t(self):
        return self.omeMetadata.image().Pixels.SizeT

    def size_x(self):
        return self.omeMetadata.image().Pixels.SizeX

    def size_y(self):
        return self.omeMetadata.image().Pixels.SizeY

    @staticmethod
    def _size_of_ndarray(data: np.ndarray) -> int:
        """
        Calculate the size of data to determine if we require bigtiff
        Returns
        -------
        the size of data in bytes
        """
        if data is None:
            return 0
        size = data.size * data.itemsize
        return size

    # set up some sensible defaults from provided info
    def _make_meta(self, data, channel_names=None, image_name="IMAGE0", pixels_physical_size=None, channel_colors=None,
                   dimension_order="TCZYX"):
        """Creates the necessary metadata for an OME tiff image

        :param data: An array of dimensions TZCYX, ZCYX, or ZYX to be written out to a file.
        :param channel_names: The names for each channel to be put into the OME metadata
        :param image_name: The name of the image to be put into the OME metadata
        :param pixels_physical_size: The physical size of each pixel in the image
        :param channel_colors: The channel colors to be put into the OME metadata
        :param dimension_order: The order of dimensions in the data array, using T,C,Z,Y and X
        """
        ox = self.omeMetadata

        ox.image().set_Name(image_name)
        ox.image().set_ID("0")
        pixels = ox.image().Pixels
        pixels.ome_uuid = ox.uuidStr
        pixels.set_ID("0")
        if pixels_physical_size is not None:
            pixels.set_PhysicalSizeX(pixels_physical_size[0])
            pixels.set_PhysicalSizeY(pixels_physical_size[1])
            pixels.set_PhysicalSizeZ(pixels_physical_size[2])

        shape = data.shape

        def dim_or_1(dim):
            idx = dimension_order.find(dim)
            return 1 if idx == -1 else shape[idx]

        pixels.channel_count = dim_or_1("C")
        pixels.set_SizeT(dim_or_1("T"))
        pixels.set_SizeC(dim_or_1("C"))
        pixels.set_SizeZ(dim_or_1("Z"))
        pixels.set_SizeY(dim_or_1("Y"))
        pixels.set_SizeX(dim_or_1("X"))

        # this must be set to the *reverse* of what dimensionality the ome tif file is saved as
        pixels.set_DimensionOrder(dimension_order[::-1])

        # convert our numpy dtype to a ome compatible pixeltype
        pixels.set_PixelType(omexml.get_pixel_type(data.dtype))

        if channel_names is None:
            for i in range(pixels.SizeC):
                pixels.Channel(i).set_ID("Channel:0:" + str(i))
                pixels.Channel(i).set_Name("C:" + str(i))
        else:
            for i in range(pixels.SizeC):
                name = channel_names[i]
                pixels.Channel(i).set_ID("Channel:0:" + str(i))
                pixels.Channel(i).set_Name(name)

        if channel_colors is not None:
            assert len(channel_colors) >= pixels.get_SizeC()
            for i in range(pixels.SizeC):
                pixels.Channel(i).set_Color(channel_colors[i])

        # assume 1 sample per channel
        for i in range(pixels.SizeC):
            pixels.Channel(i).set_SamplesPerPixel(1)

        # many assumptions in here: one file per image, one plane per tiffdata, etc.
        pixels.populate_TiffData()

        return ox



import io
import logging
import re

import numpy as np
import tifffile
from typing import Tuple

from .. import types
from ..vendor import omexml
from .reader import Reader
from .tiff_reader import TiffReader

log = logging.getLogger(__name__)


class OmeTiffReader(Reader):
    """Opening and processing the contents of an OME Tiff file
    """

    def __init__(self, file: types.FileLike, **kwargs):
        super().__init__(file, **kwargs)
        try:
            self.tiff = tifffile.TiffFile(self._bytes)
        except Exception:
            log.error("tiffile could not parse this input")
            raise

    def _lazy_init_metadata(self) -> omexml.OMEXML:
        if self._metadata is None and self.tiff.is_ome:
            description = self.tiff.pages[0].description.strip()
            if not (description.startswith("<?xml version=") and description.endswith("</OME>")):
                raise ValueError(f'Description does not conform to OME specification: {description[:100]}')
            self._metadata = omexml.OMEXML(description)
        return self._metadata

    @staticmethod
    def _is_this_type(buffer: io.BufferedIOBase) -> bool:
        is_tif = TiffReader._is_this_type(buffer)
        if is_tif:
            buf = TiffReader.get_image_description(buffer)
            if buf is None:
                return False
            if buf[0:5] != b"<?xml":
                return False
            match = re.search(
                b'<(\\w*)(:?)OME [^>]*xmlns\\2\\1="http://www.openmicroscopy.org/Schemas/[Oo][Mm][Ee]/',
                buf
            )
            if match is None:
                return False
            return True
        return False

    @property
    def data(self) -> np.ndarray:
        if self._data is None:
            # load the data
            self._data = self.tiff.asarray()
        return self._data

    @property
    def dims(self) -> str:
        self._lazy_init_metadata()
        dimension_order = self._metadata.image().Pixels.DimensionOrder
        # reverse the string
        dimension_order = dimension_order[::-1]
        # see if t,z,or c is squeezed out.
        # this is a tifffile implementation detail -- see squeeze_axes in tifffile.
        if self.size_t() < 2:
            dimension_order = dimension_order.replace("T", "")
        if self.size_z() < 2:
            dimension_order = dimension_order.replace("Z", "")
        if self.size_c() < 2:
            dimension_order = dimension_order.replace("C", "")
        return dimension_order

    @property
    def metadata(self) -> omexml.OMEXML:
        return self._lazy_init_metadata()

    def load_slice(self, slice_index=0):
        data = self.tiff.asarray(key=slice_index)
        return data

    def size_z(self):
        return self.metadata.image().Pixels.SizeZ

    def size_c(self):
        return self.metadata.image().Pixels.SizeC

    def size_t(self):
        return self.metadata.image().Pixels.SizeT

    def size_x(self):
        return self.metadata.image().Pixels.SizeX

    def size_y(self):
        return self.metadata.image().Pixels.SizeY

    def dtype(self):
        return self.tiff.pages[0].dtype

    def is_ome(self):
        return OmeTiffReader._is_this_type(self._bytes)

    def get_channel_names(self, scene: int = 0):
        return self.metadata.image(scene).Pixels.get_channel_names()

    def get_physical_pixel_size(self, scene: int = 0) -> Tuple[float]:
        p = self.metadata.image(scene).Pixels
        return (p.get_PhysicalSizeX(), p.get_PhysicalSizeY(), p.get_PhysicalSizeZ())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "numpy>=1.16",
    "imageio>=2.3.0",
    "scipy>=1.1.0",
    "tifffile==0.15.1",
]

test_requirements = [
    "codecov",
    "flake8",
    "pytest",
    "pytest-cov",
    "pytest-raises",
]

dev_requirements = [
    "bumpversion>=0.5.3",
    "coverage>=5.0a4",
    "flake8>=3.7.7",
    "ipython>=7.5.0",
    "m2r>=0.2.1",
    "pytest>=4.3.0",
    "pytest-cov==2.6.1",
    "pytest-raises>=0.10",
    "pytest-runner>=4.4",
    "Sphinx>=2.0.0b1",
    "sphinx_rtd_theme>=0.1.2",
    "tox>=3.5.2",
    "twine>=1.13.0",
    "wheel>=0.33.1",
]

setup_requirements = [
    "pytest-runner",
]

interactive_requirements = [
    "altair",
    "jupyterlab",
    "matplotlib",
    "pillow",
]

extra_requirements = {
    "test": test_requirements,
    "dev": dev_requirements,
    "setup": setup_requirements,
    "interactive": interactive_requirements,
    "all": [
        *requirements,
        *test_requirements,
        *setup_requirements,
        *dev_requirements,
        *interactive_requirements
    ]
}

setup(
    author="Allen Institute for Cell Science",
    author_email=(
        "jacksonb@alleninstitute.org, "
        "bowdenm@spu.edu"
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description=(
        "Python library for reading and writing image data with special handlers for bio-formats "
        "from Allen Institute for Cell Science."
    ),
    entry_points={},
    install_requires=requirements,
    license="BSD-3-Clause",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="aicsimageio, allen cell, imaging, computational biology",
    name="aicsimageio",
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*"]),
    python_requires=">=3.6",
    setup_requires=setup_requirements,
    test_suite="aicsimageio/tests",
    tests_require=test_requirements,
    extras_require=extra_requirements,
    url="https://github.com/AllenCellModeling/aicsimageio",
    # Do not edit this string manually, always use bumpversion
    # Details in CONTRIBUTING.md
    version="3.0.7",
    zip_safe=False,
)
